@@ -47,26 +47,10 @@ class ByteReader(object):
         self.pos = 0
         self._eof = EOFError('unexpected end of file')
 
-    def _print_bytes(self, n):
-        """Print the next n bytes as hex and characters.
-
-        This is used for debugging.
-        """
-        data = self._buffer[self.pos:self.pos + n]
-        for pos, byte in enumerate(data, start=self.pos):
-            char = chr(byte)
-            if not char in string.printable or char in string.whitespace:
-                char = ''
-            print('  {:06x}: {:02x} {}'.format(pos, byte, char))
-
-        if len(data) < n:
-            raise EOFError('end of file reached in read_list()')
-
     def read_byte(self):
         """Read one byte."""
         try:
             byte = self._buffer[self.pos]
-            # self._print_bytes(1)
             self.pos += 1
             return byte
         except IndexError:
@@ -81,16 +65,9 @@ class ByteReader(object):
         except IndexError:
             raise self._eof
 
-    def read_list(self, n):
-        """Read n bytes and return as a list."""
-        # self._print_bytes(n)
-        i = self.pos
-        ret = self._buffer[i:i + n]
-        if len(ret) < n:
-            raise self._eof
-
-        self.pos += n
-        return ret
+    def peek_list(self, n):
+        """Return a list of the next n bytes."""
+        return self._buffer[self.pos:self.pos+n]
 
     def read_short(self):
         """Read short (2 bytes little endian)."""
@@ -102,6 +79,16 @@ class ByteReader(object):
         a, b, c, d = self.read_list(4)
         return a << 24 | b << 16 | c << 8 | d
 
+    def read_list(self, n):
+        """Read n bytes and return as a list."""
+        i = self.pos
+        ret = self._buffer[i:i + n]
+        if len(ret) < n:
+            raise self._eof
+
+        self.pos += n
+        return ret
+
     def __enter__(self):
         return self
 
@@ -206,9 +193,10 @@ class MidiFile:
         with ByteReader(self.filename) as self._file, \
                 meta_charset(self.charset):
             # Read header (16 bytes)
-            magic = self._file.read_list(4)
+            magic = self._file.peek_list(4)
             if not bytearray(magic) == bytearray(b'MThd'):
                 raise IOError('MThd not found. Probably not a MIDI file')
+            self._file.read_list(4)  # Skip MThd
 
             # This is always 6 for any file created under the MIDI 1.0
             # specification, but this allows for future expansion.
@@ -243,7 +231,7 @@ class MidiFile:
         data_bytes = self._file.read_list(spec.length - 1)
         for byte in data_bytes:
             if byte > 127:
-                raise IOError('data byte has value > 127')
+                raise IOError('data byte must be in range 0..127')
         return build_message(spec, [status_byte] + data_bytes)
 
     def _read_meta_message(self):
@@ -253,9 +241,12 @@ class MidiFile:
         return _build_meta_message(type, data)
 
     def _read_sysex(self):
-        length = self._file.read_byte()
+        length = self._read_variable_int()
         data = self._file.read_list(length)
 
+        # Strip start and end bytes.
+        if data and data[0] == 0xf0:
+            data = data[1:]
         if data and data[-1] == 0xf7:
             data = data[:-1]
 
@@ -265,22 +256,25 @@ class MidiFile:
     def _read_track(self):
         track = MidiTrack()
 
-        magic = self._file.read_list(4)
-        if bytearray(magic) != bytearray(b'MTrk'):
-            raise IOError("track doesn't start with 'MTrk'")
+        magic = self._file.peek_list(4)
+        if bytearray(magic) == bytearray(b'MTrk'):
+            self._file.read_list(4)  # Skip 'MTrk'
+            length = self._file.read_long()
+        else:
+            # Todo: some files don't have track headers?
+            # These end with end_of_track or end of file,
+            # so we set length to infinite.
+            # raise IOError("track doesn't start with 'MTrk'")
+            length = float('inf')
 
-        length = self._file.read_long()
         start = self._file.pos
         last_status = None
-        last_sysex = None
 
         while 1:
             # End of track reached.
             if self._file.pos - start == length:
                 break
 
-            # print('--- New')
-
             delta = self._read_variable_int()
 
             # Todo: not all messages have running status
@@ -297,24 +291,16 @@ class MidiFile:
 
             if status_byte == 0xff:
                 message = self._read_meta_message()
-            elif status_byte == 0xf0:
+            elif status_byte in [0xf0, 0xf7]:
+                # Todo: I'm not quite clear on the difference between
+                # f0 and f7 events.
                 message = self._read_sysex()
-                last_sysex = message
-            elif status_byte == 0xf7:
-                # Todo: check if this works as intended.
-                if last_sysex is None:
-                    raise IOError(
-                        'sysex continuation without preceding sysex')
-                last_sysex.data += self._read_sysex()
-                continue
             else:
                 message = self._read_message(status_byte)
 
             message.time = delta
             track.append(message)
 
-            # print('---', message)
-
             if message.type == 'end_of_track':
                 break
 
@@ -463,7 +449,14 @@ class MidiFile:
 
                     # Todo: running status?
                     bytes += encode_variable_int(message.time)
-                    bytes += message.bytes()
+                    if message.type == 'sysex':
+                        bytes += [0xf0]
+                        # length (+ 1 for end byte (0xf7))
+                        bytes += encode_variable_int(len(message.data) + 1)
+                        bytes += message.data
+                        bytes += [0xf7]
+                    else:
+                        bytes += message.bytes()
 
                 if not self._has_end_of_track(track):
                     # Write end_of_track.
@@ -501,3 +494,70 @@ class MidiFile:
 
     def __exit__(self, type, value, traceback):
         return False
+
+class _DebugByteReader(ByteReader):
+    parent = ByteReader
+
+    def _print_bytes(self, n):
+        """Print the next n bytes as hex and characters.
+
+        This is used for debugging.
+        """
+        data = self._buffer[self.pos:self.pos + n]
+        print()
+        for pos, byte in enumerate(data, start=self.pos):
+            char = chr(byte)
+            if not char in string.printable or char in string.whitespace:
+                char = ''
+            print('  {:06x}: {:02x} {}'.format(pos, byte, char))
+
+        if len(data) < n:
+            raise EOFError('unexpected end of file')
+
+    def read_byte(self):
+        self._print_bytes(1)
+        return self.parent.read_byte(self)
+
+    def read_short(self):
+        self._print_bytes(2)
+        return self.parent.read_short(self)
+
+    def read_long(self):
+        self._print_bytes(4)
+        return self.parent.read_long(self)
+
+    def read_list(self, n):
+        self._print_bytes(n)
+        return self.parent.read_list(self, n)
+
+class _DebugMidiFile(MidiFile):
+    parent = MidiFile
+
+    def _read_track(self):
+        print('-- Track', len(self.tracks))
+        return self.parent._read_track(self)
+
+    def _read_message(self, status_byte):
+        print('( new message:')
+        message = self.parent._read_message(self, status_byte)
+        print(')', message)
+        return message
+
+    def _read_meta_message(self):
+        print('( new meta message')
+        message = self.parent._read_meta_message(self)
+        print(')', message)
+        return message
+
+    def _read_sysex(self):
+        print('( new sysex message')
+        message = self.parent._read_sysex(self)
+        print(')', message)
+        return message
+
+def debug():
+    """Turn on debugging prints."""
+    global ByteReader, MidiFile
+
+    ByteReader = _DebugByteReader
+    MidiFile = _DebugMidiFile
